/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004-2005 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

#ifdef __EMX__
#include <strings.h>
#endif 

#include <bro_types.h>
#include <bro_val.h>
#include <bro_debug.h>
#include <bro_object.h>
#include <bro_type.h>


BroType *
__bro_type_new(void)
{
  BroType *type;

  D_ENTER;

  if (! (type = calloc(1, sizeof(BroType))))
    D_RETURN_(NULL);

  __bro_type_init(type);

  D_RETURN_(type);
}


BroType *
__bro_type_new_of_type(int type_tag, const char *type_name)
{
  BroType *type = NULL;
  int internal_tag;
  char is_nbo = 0;
  
  D_ENTER;
  
  switch (type_tag)
    {
    case BRO_TYPE_BOOL:
    case BRO_TYPE_INT:
    case BRO_TYPE_ENUM:
      internal_tag = BRO_INTTYPE_INT;
      break;
      
    case BRO_TYPE_COUNT:
    case BRO_TYPE_COUNTER:
      internal_tag = BRO_INTTYPE_UNSIGNED;
      break;
      
    case BRO_TYPE_PORT:
      internal_tag = BRO_INTTYPE_UNSIGNED;
      is_nbo = 1;
      break;
      
    case BRO_TYPE_DOUBLE:
    case BRO_TYPE_TIME:
    case BRO_TYPE_INTERVAL:
      internal_tag = BRO_INTTYPE_DOUBLE;
      break;
      
    case BRO_TYPE_STRING:
      internal_tag = BRO_INTTYPE_STRING;
      break;
      
    case BRO_TYPE_IPADDR:
    case BRO_TYPE_NET:
      internal_tag = BRO_INTTYPE_IPADDR;
      break;
      
    case BRO_TYPE_SUBNET:
      internal_tag = BRO_INTTYPE_SUBNET;
      break;
      
    case BRO_TYPE_PATTERN:
    case BRO_TYPE_TIMER:
    case BRO_TYPE_ANY:
    case BRO_TYPE_TABLE:
    case BRO_TYPE_UNION:
    case BRO_TYPE_LIST:      
    case BRO_TYPE_FUNC:
    case BRO_TYPE_FILE:
    case BRO_TYPE_VECTOR:
      internal_tag = BRO_INTTYPE_OTHER;
      break;

    case BRO_TYPE_RECORD:
      if (! (type = (BroType *) __bro_record_type_new()))
	D_RETURN_(NULL);
      
      internal_tag = BRO_INTTYPE_OTHER;
      break;
      
    case BRO_TYPE_ERROR:
    default:
      internal_tag = BRO_INTTYPE_ERROR;
    }

  if (! type)
    {
      if (! (type = __bro_type_new()))
	D_RETURN_(NULL);
    }
 
  type->tag          = type_tag;
  type->internal_tag = internal_tag;
  type->is_nbo       = is_nbo;
  type->is_complete  = TRUE;
  
  if (type_name)
    {
      type->is_complete = FALSE;
      bro_string_set(&type->type_name, type_name);
    }

  D_RETURN_(type);
}


void
__bro_type_init(BroType *type)
{
  BroSObject *sobj = (BroSObject *) type;

  D_ENTER;

  __bro_object_init((BroObject *) type);

  sobj->read  = (BroSObjectRead) __bro_type_read;
  sobj->write = (BroSObjectWrite) __bro_type_write;
  sobj->free  = (BroSObjectFree) __bro_type_free;
  sobj->clone = (BroSObjectClone) __bro_type_clone;

  sobj->type_id = SER_TYPE;

  bro_string_init(&type->type_name);
  type->is_complete = TRUE;

  D_RETURN;
}


void
__bro_type_free(BroType *type)
{
  D_ENTER;
  bro_string_cleanup(&type->type_name);
  __bro_sobject_release((BroSObject *) type->attrs_type);
  __bro_object_free((BroObject *) type);
  D_RETURN;
}


int
__bro_type_read(BroType *type, BroConn *bc)
{
  char opt;
  
  D_ENTER;

  if (! __bro_object_read((BroObject *) type, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_read_char(bc->rx_buf, &type->tag))
    D_RETURN_(FALSE);
  if (! __bro_buf_read_char(bc->rx_buf, &type->internal_tag))
    D_RETURN_(FALSE);

  if (! __bro_buf_read_char(bc->rx_buf, &type->is_nbo))
    D_RETURN_(FALSE);
  if (! __bro_buf_read_char(bc->rx_buf, &type->is_base_type))
    D_RETURN_(FALSE);
  if (! __bro_buf_read_char(bc->rx_buf, &type->is_global_attrs_type))
    D_RETURN_(FALSE);

  if (! __bro_buf_read_char(bc->rx_buf, &opt))
    D_RETURN_(FALSE);
  if (opt)
    {
      if (type->attrs_type)
	__bro_sobject_release((BroSObject *) type->attrs_type);
      
      if (! (type->attrs_type = (BroRecordType *)
	     __bro_sobject_unserialize(SER_RECORD_TYPE, bc)))
	D_RETURN_(FALSE);
    }

  D_RETURN_(TRUE);
}


int
__bro_type_write(BroType *type, BroConn *bc)
{
  D_ENTER;

   if (! __bro_object_write((BroObject *) type, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_char(bc->tx_buf, type->tag))
    D_RETURN_(FALSE);
  if (! __bro_buf_write_char(bc->tx_buf, type->internal_tag))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_char(bc->tx_buf, type->is_nbo))
    D_RETURN_(FALSE);
  if (! __bro_buf_write_char(bc->tx_buf, type->is_base_type))
    D_RETURN_(FALSE);
  if (! __bro_buf_write_char(bc->tx_buf, type->is_global_attrs_type))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_char(bc->tx_buf, type->attrs_type ? 1 : 0))
    D_RETURN_(FALSE);
  
  if (type->attrs_type && ! __bro_sobject_serialize((BroSObject *) type->attrs_type, bc))
    D_RETURN_(FALSE);
  
  D_RETURN_(TRUE);
}


int
__bro_type_clone(BroType *dst, BroType *src)
{
  D_ENTER;

  if (! __bro_object_clone((BroObject *) dst, (BroObject *) src))
    D_RETURN_(FALSE);

  dst->tag = src->tag;
  dst->internal_tag = src->internal_tag;
  dst->is_nbo = src->is_nbo;
  dst->is_base_type = src->is_base_type;
  dst->is_global_attrs_type = src->is_global_attrs_type;
  dst->is_complete = src->is_complete;
  bro_string_set(&dst->type_name, bro_string_get_data(&src->type_name));

  if (src->attrs_type &&
      ! (dst->attrs_type = (BroRecordType *) __bro_sobject_copy((BroSObject *) src->attrs_type)))
    D_RETURN_(FALSE);

  D_RETURN_(TRUE);
}


BroTypeList *
__bro_type_list_new(void)
{
  BroTypeList *tl;

  D_ENTER;

  if (! (tl = calloc(1, sizeof(BroTypeList))))
    D_RETURN_(NULL);

  __bro_type_list_init(tl);

  D_RETURN_(tl);
}


void
__bro_type_list_init(BroTypeList *tl)
{
  BroSObject *sobj = (BroSObject *) tl;

  D_ENTER;
  
  __bro_object_init((BroObject *) tl);
  
  sobj->read  = (BroSObjectRead) __bro_type_list_read;
  sobj->write = (BroSObjectWrite) __bro_type_list_write;
  sobj->free  = (BroSObjectFree) __bro_type_list_free;
  sobj->clone  = (BroSObjectClone) __bro_type_list_clone;

  sobj->type_id = SER_TYPE_LIST;
  
  tl->types = NULL;

  D_RETURN;
}


void
__bro_type_list_free(BroTypeList *tl)
{
  D_ENTER;

  __bro_list_free(tl->types, (BroFunc) __bro_sobject_release);
  __bro_sobject_release((BroSObject *) tl->pure_type);  
  __bro_object_free((BroObject *) tl);
  
  D_RETURN;
}


int
__bro_type_list_read(BroTypeList *tl, BroConn *bc)
{
  char opt;
  uint32 i;

  D_ENTER;

  if (! __bro_object_read((BroObject *) tl, bc))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_read_char(bc->rx_buf, &opt))
    D_RETURN_(FALSE);
  
  /* Clean out old optional pure type */
  if (tl->pure_type)
    __bro_sobject_release((BroSObject *) tl->pure_type);
  
  if (opt)
    {
      if (! (tl->pure_type = (BroType *) __bro_sobject_unserialize(SER_IS_TYPE, bc)))
	D_RETURN_(FALSE);
    }
  
  if (! __bro_buf_read_int(bc->rx_buf, &tl->num_types))
    D_RETURN_(FALSE);
  
  if (tl->num_types > 0)
    {
      for (i = 0; i < tl->num_types; i++)
	{
	  BroType *type;
	  
	  if (! (type = (BroType *) __bro_sobject_unserialize(SER_IS_TYPE, bc)))
	    D_RETURN_(FALSE);
	  
	  tl->types = __bro_list_append(tl->types, type);
	}
    }
  
  D_RETURN_(TRUE);
}


int
__bro_type_list_write(BroTypeList *tl, BroConn *bc)
{
  BroList *l;

  D_ENTER;

  if (! __bro_object_write((BroObject *) tl, bc))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_write_char(bc->tx_buf, tl->pure_type ? 1 : 0))
    D_RETURN_(FALSE);
  
  if (tl->pure_type && ! __bro_sobject_serialize((BroSObject *) tl->pure_type, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_int(bc->tx_buf, tl->num_types))
    D_RETURN_(FALSE);

  for (l = tl->types; l; l = __bro_list_next(l))
    {
      if (! __bro_sobject_serialize((BroSObject *) __bro_list_data(l), bc))
	D_RETURN_(FALSE);
    }
  
  D_RETURN_(TRUE);
}


int
__bro_type_list_clone(BroTypeList *dst, BroTypeList *src)
{
  BroList *l;
  BroType *type;

  D_ENTER;

  if (! __bro_object_clone((BroObject *) dst, (BroObject *) src))
    D_RETURN_(FALSE);

  dst->num_types = src->num_types;

  if (dst->types)
    __bro_list_free(dst->types, (BroFunc) __bro_sobject_release);

  dst->types = NULL;
  
  for (l = src->types; l; l = __bro_list_next(l))
    {
      BroType *type_copy;

      type = __bro_list_data(l);
      
      if (! (type_copy = (BroType *) __bro_sobject_copy((BroSObject *) type)))
	D_RETURN_(FALSE);

      dst->types = __bro_list_append(dst->types, type_copy);
    }
  
  if (src->pure_type && ! (dst->pure_type = (BroType *) __bro_sobject_copy((BroSObject *) src->pure_type)))
    D_RETURN_(FALSE);
  
  D_RETURN_(TRUE);
}


BroTypeDecl *
__bro_type_decl_new(void)
{
  BroTypeDecl *td;

  D_ENTER;

  if (! (td = calloc(1, sizeof(BroTypeDecl))))
    D_RETURN_(NULL);

  D_RETURN_(td);
}


BroTypeDecl     *
__bro_type_decl_copy(BroTypeDecl *td)
{
  BroTypeDecl *copy;

  D_ENTER;

  if (! td)
    D_RETURN_(NULL);

  if (! (copy = __bro_type_decl_new()))
    D_RETURN_(NULL);

  if (td->attrs && ! (copy->attrs = (BroAttrs *) __bro_sobject_copy((BroSObject *) td->attrs)))
    goto error_result;

  if (td->type && ! (copy->type = (BroType *) __bro_sobject_copy((BroSObject *) td->type)))
    goto error_result;
  
  if (! (bro_string_set_data(&copy->id, bro_string_get_data(&td->id),
			     bro_string_get_length(&td->id))))
    goto error_result;
  
  D_RETURN_(copy);
  
 error_result:
  __bro_type_decl_free(copy);
  D_RETURN_(NULL);
}


void
__bro_type_decl_free(BroTypeDecl *td)
{
  D_ENTER;

  if (! td)
    D_RETURN;

  __bro_sobject_release((BroSObject *) td->type);
  __bro_sobject_release((BroSObject *) td->attrs);
  bro_string_cleanup(&td->id);
  free(td);
  
  D_RETURN;
}


int
__bro_type_decl_read(BroTypeDecl *td, BroConn *bc)
{
  char opt;

  D_ENTER;

  if (! td || !bc)
    D_RETURN_(FALSE);

  /* Read an optional BroAttrs */

  if (td->attrs)
    __bro_sobject_release((BroSObject *) td->attrs);
  td->attrs = NULL;

  if (! __bro_buf_read_char(bc->rx_buf, &opt))
    D_RETURN_(FALSE);
  if (opt)
    {
      if (! (td->attrs = (BroAttrs *) __bro_sobject_unserialize(SER_ATTRIBUTES, bc)))
	D_RETURN_(FALSE);
    }

  /* Read a type */

  if (td->type)
    __bro_sobject_release((BroSObject *) td->type);
  td->type = NULL;

  if (! (td->type = (BroType *) __bro_sobject_unserialize(SER_IS_TYPE, bc)))
    D_RETURN_(FALSE);

  /* Read ID name string */
  
  bro_string_cleanup(&td->id);
  if (! __bro_buf_read_string(bc->rx_buf, &td->id))
    D_RETURN_(FALSE);

  D_RETURN_(TRUE);
}


int
__bro_type_decl_write(BroTypeDecl *td, BroConn *bc)
{
  D_ENTER;

  if (! td || !bc)
    D_RETURN_(FALSE);

  if (! __bro_buf_write_char(bc->tx_buf, td->attrs ? 1 : 0))
    D_RETURN_(FALSE);

  if (td->attrs && ! __bro_sobject_serialize((BroSObject *) td->attrs, bc))
    D_RETURN_(FALSE);
  
  if (! __bro_sobject_serialize((BroSObject *) td->type, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_string(bc->tx_buf, &td->id))
    D_RETURN_(FALSE);
  
  D_RETURN_(TRUE);
}


BroRecordType *
__bro_record_type_new(void)
{
  BroRecordType *rt;
  
  D_ENTER;
  
  if (! (rt = calloc(1, sizeof(BroRecordType))))
    D_RETURN_(NULL);
  
  __bro_record_type_init(rt);

  D_RETURN_(rt);
}


void
__bro_record_type_init(BroRecordType *rt)
{
  BroSObject *sobj = (BroSObject *) rt;

  D_ENTER;
  
  __bro_type_init((BroType *) rt);
  
  sobj->read  = (BroSObjectRead) __bro_record_type_read;
  sobj->write = (BroSObjectWrite) __bro_record_type_write;
  sobj->free  = (BroSObjectFree) __bro_record_type_free;
  sobj->clone = (BroSObjectClone) __bro_record_type_clone;
  
  sobj->type_id = SER_RECORD_TYPE;
  
  D_RETURN;
}


void
__bro_record_type_free(BroRecordType *rt)
{
  D_ENTER;
  
  __bro_sobject_release((BroSObject *) rt->base);
  __bro_list_free(rt->type_decls, (BroFunc) __bro_type_decl_free);
  __bro_type_free((BroType *) rt);
  
  D_RETURN;
}


void
__bro_record_type_add_type(BroRecordType *rt, const char *field, BroType *type)
{
  BroTypeDecl *td;
  
  D_ENTER;
  
  if (! rt || ! type)
    D_RETURN;
  
  if (! (td = __bro_type_decl_new()))
    D_RETURN;

  if (! (td->type = (BroType *) __bro_sobject_copy((BroSObject *) type)))
    {
      D(("Cloning of type failed.\n"));
      __bro_type_decl_free(td);
      D_RETURN;
    }
  
  if (! bro_string_set(&td->id, field))
    {
      __bro_type_decl_free(td);
      D_RETURN;
    }
  
  rt->type_decls = __bro_list_append(rt->type_decls, td);
  rt->num_fields++;
  rt->num_types++;
  D_RETURN;
}


const char *
__bro_record_type_get_nth_field(BroRecordType *rt, int num)
{
  BroList *l;
  
  if (! rt || num < 0 || (uint) num >= rt->num_fields)
    return NULL;
  
  if( (l = __bro_list_nth(rt->type_decls, num)))
    {
      BroTypeDecl *td = __bro_list_data(l);
      return td->id.str_val;
    }
  
  return NULL;
}


int
__bro_record_type_read(BroRecordType *rt, BroConn *bc)
{
  uint32 i;
  char opt;

  D_ENTER;

  if (! __bro_type_read((BroType *) rt, bc))
    D_RETURN_(FALSE);

  /* Read type declarations */

  rt->type_decls = NULL;
  rt->num_fields = 0;
  rt->num_types = 0;

  if (! __bro_buf_read_int(bc->rx_buf, &rt->num_fields))
    D_RETURN_(FALSE);

  if (! __bro_buf_read_char(bc->rx_buf, &opt))
    D_RETURN_(FALSE);
  if (opt)
    {
      if (! __bro_buf_read_int(bc->rx_buf, &rt->num_types))
	D_RETURN_(FALSE);
      
      if (rt->num_types > 0)
	{
	  for (i = 0; i < rt->num_types; i++)
	    {
	      BroTypeDecl *td;
	      
	      if (! (td = __bro_type_decl_new()))
		D_RETURN_(FALSE);
	      
	      if (! __bro_type_decl_read(td, bc))
		D_RETURN_(FALSE);

	      rt->type_decls = __bro_list_append(rt->type_decls, td);
	    }
	}
    }

  /* Read optional base */

  __bro_sobject_release((BroSObject *) rt->base);
  rt->base = NULL;
  
  if (! __bro_buf_read_char(bc->rx_buf, &opt))
    D_RETURN_(FALSE);
  if (opt)
    {
      if (! (rt->base = (BroTypeList *) __bro_sobject_unserialize(SER_TYPE_LIST, bc)))
	D_RETURN_(FALSE);
    }
  
  D_RETURN_(TRUE);
}


int
__bro_record_type_write(BroRecordType *rt, BroConn *bc)
{
  BroList *l;

  D_ENTER;

  if (! __bro_type_write((BroType *) rt, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_int(bc->tx_buf, rt->num_fields))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_char(bc->tx_buf, rt->type_decls ? 1 : 0))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_write_int(bc->tx_buf, rt->num_types))
    D_RETURN_(FALSE);
  
  for (l = rt->type_decls; l; l = __bro_list_next(l))
    {
      BroTypeDecl *td = __bro_list_data(l);
      
      if (! __bro_type_decl_write(td, bc))
	D_RETURN_(FALSE);
    }
  
  if (! __bro_buf_write_char(bc->tx_buf, rt->base ? 1 : 0))
    D_RETURN_(FALSE);
  
  if (rt->base && ! __bro_sobject_serialize((BroSObject *) rt->base, bc))
    D_RETURN_(FALSE);
  
  D_RETURN_(TRUE);
}


int
__bro_record_type_clone(BroRecordType *dst, BroRecordType *src)
{
  BroList *l;

  D_ENTER;

  if (! __bro_type_clone((BroType *) dst, (BroType *) src))
    D_RETURN_(FALSE);
  
  if (src->base && ! (dst->base = (BroTypeList *) __bro_sobject_copy((BroSObject *) src->base)))
    D_RETURN_(FALSE);
  
  dst->num_fields = src->num_fields;
  dst->num_types = src->num_types;
  
  for (l = src->type_decls; l; l = __bro_list_next(l))
    {
      BroTypeDecl *td = __bro_list_data(l);
      BroTypeDecl *td_copy = __bro_type_decl_copy(td);
      
      if (! td_copy)
	D_RETURN_(FALSE);
 
      dst->type_decls = __bro_list_append(dst->type_decls, td_copy);
    }
  
  D_RETURN_(TRUE);
}
